
// Embed an array of virtual processes ``optimally'' within a given 
// number of physical processes.

//   void mapVirtuals(int q [], int n [], int r, int p) ;
//
//     p = number of available physical process
//     r = rank of virtual process array
//     n = r-element shape vector for virtual process array
//     q = r-element shape vector for physical process array (returned)

// The product of the elements of `q' is not greater than `p', and
// blockwise distribution of the shape-n virtual array in a shape-q
// process array gives optimal ``load balancing'' (defined to minimise the
// maximum number of virtual processes on any physical process).

// This implementation uses a brute-force search of all r-tuples for
// `q' (whose product is not greater than `p').  It is not suitable
// for very large values of `p', or moderately large values of
// `r'.  (For fixed `r', complexity is polynomially bounded in `p'.
// For fixed `p', complexity is exponential in `r'.)

// The code is amenable to extensive optimisation, by eliminating
// recursion, and repeated computations of products, etc.


// Global variables...

static int P ;     // Copy of `p'.
static int* N ;    // Copy of `n'.
static int* Q ;    // Current best estimate for `q'...
static int load ;  // ...corresponding max virtual process ``load''.


static void operate(int i [], int r) {
  // `i' is an r-tuple---the current trial for `q' is had by
  // incrementing the elements of `i' by 1.

  // load is defined as product of [(n_i + q_i - 1) / q_i].

  int d, loadi ;

  loadi = 1 ;
  for(d = 0 ; d < r ; d++)
    loadi *= (i [d] + N [d]) / (i [d] + 1) ;

  if(loadi < load) {
    load = loadi ;
    for(d = 0 ; d < r ; d++)
      Q [d] = i [d] + 1 ;
  }
}


// Loop test: product of i + 1's <= P

static int test(int d, int i []) {
  int e, prod = 1 ;

  for(e = 0 ; e <= d ; e++)
    prod *= i [e] + 1 ;

  return prod <= P ;
}


// Fairly general r-dimensional loop...

static void iterate(int d, int i [], int r) {
  if(d == r)
    operate(i, r) ;
  else
    for(i [d] = 0 ; test(d, i) ; i [d] ++)
      iterate(d + 1, i, r) ;
}


void mapVirtuals(int q [], int n [], int r, int p) {
  int d ;
  N = new int [r] ;
  for(d = 0 ; d < r ; d++)
    N [d] = n [d] ;
  Q = new int [r] ;
  P = p ;

  // Worst case---sequential single-processor mapping.
  load = 1 ;
  for(d = 0 ; d < r ; d++) {
    Q [d] = 1 ;
    load *= N [d] ;
  }

  // Improve mapping.
  int* i = new int [r] ;
  iterate(0, i, r) ;
  delete i ;

  for(d = 0 ; d < r ; d++)
    q [d] = Q [d] ;
  delete Q ;
  delete N ;
}

// // test program
//
// #include <iostream.h>
//
// int main() {
//   int r, d, p ;
// 
//   // Get shape of virtual process array
// 
//   cout << "r?: " ;
//   cin  >> r ;
// 
//   int* n = new int [r] ;
// 
//   for(d = 0 ; d < r ; d++) {
//     cout << "n_" << d + 1 << "?: " ;
//     cin  >> n [d] ;
//   }
// 
//   // Get number of physical processes
// 
//   cout << "p?: " ;
//   cin >> p ;
// 
//   int* q = new int [r] ;
// 
//   mapVirtuals(q, n, r, p) ;
// 
//   for(d = 0 ; d < r ; d++)
//     cout << "p_" << d + 1 << " = " << q [d] << "\n" ;
//   delete n ;
// }
// 
// 
